@@ -1,8 +1,8 @@
 /******************************************************************************
- * $Id: gdal_pam.h 21291 2010-12-19 20:42:06Z rouault $
+ * $Id: gdal_pam.h 2e6733101e6356653e8c999cb7fe808abc22ed11 2020-08-11 14:57:51 +0200 Even Rouault $
  *
  * Project:  GDAL Core
- * Purpose:  Declaration for Peristable Auxilary Metadata classes.
+ * Purpose:  Declaration for Peristable Auxiliary Metadata classes.
  * Author:   Frank Warmerdam, warmerdam@pobox.com
  *
  ******************************************************************************
@@ -30,7 +30,10 @@
 #ifndef GDAL_PAM_H_INCLUDED
 #define GDAL_PAM_H_INCLUDED
 
+//! @cond Doxygen_Suppress
+
 #include "gdal_priv.h"
+#include <map>
 
 class GDALPamRasterBand;
 
@@ -64,19 +67,20 @@ class GDALPamRasterBand;
                                  GCIF_COLORINTERP | GCIF_BAND_METADATA |   \
                                  GCIF_RAT | GCIF_MASK |                    \
                                  GCIF_ONLY_IF_MISSING | GCIF_PROCESS_BANDS|\
-                                 GCIF_BAND_DESCRIPTION) 
+                                 GCIF_BAND_DESCRIPTION)
 
 /* GDAL PAM Flags */
-#define GPF_DIRTY		0x01  // .pam file needs to be written on close
+/* ERO 2011/04/13 : GPF_AUXMODE seems to be unimplemented */
+#define GPF_DIRTY               0x01  // .pam file needs to be written on close
 #define GPF_TRIED_READ_FAILED   0x02  // no need to keep trying to read .pam.
-#define GPF_DISABLED            0x04  // do not try any PAM stuff. 
+#define GPF_DISABLED            0x04  // do not try any PAM stuff.
 #define GPF_AUXMODE             0x08  // store info in .aux (HFA) file.
 #define GPF_NOSAVE              0x10  // do not try to save pam info.
 
 /* ==================================================================== */
 /*      GDALDatasetPamInfo                                              */
 /*                                                                      */
-/*      We make these things a seperate structure of information        */
+/*      We make these things a separate structure of information        */
 /*      primarily so we can modify it without altering the size of      */
 /*      the GDALPamDataset.  It is an effort to reduce ABI churn for    */
 /*      driver plugins.                                                 */
@@ -84,45 +88,72 @@ class GDALPamRasterBand;
 class GDALDatasetPamInfo
 {
 public:
-    char        *pszPamFilename;
+    char        *pszPamFilename = nullptr;
 
-    char	*pszProjection;
+    OGRSpatialReference* poSRS = nullptr;
 
-    int         bHaveGeoTransform;
-    double      adfGeoTransform[6];
+    int         bHaveGeoTransform = false;
+    double      adfGeoTransform[6]{0,0,0,0,0,0};
 
-    int         nGCPCount;
-    GDAL_GCP   *pasGCPList;
-    char       *pszGCPProjection;
+    int         nGCPCount = 0;
+    GDAL_GCP   *pasGCPList = nullptr;
+    OGRSpatialReference* poGCP_SRS = nullptr;
 
-    CPLString   osPhysicalFilename;
-    CPLString   osSubdatasetName;
-    CPLString   osAuxFilename;
+    CPLString   osPhysicalFilename{};
+    CPLString   osSubdatasetName{};
+    CPLString   osAuxFilename{};
+
+    int         bHasMetadata = false;
+
+    struct Statistics
+    {
+        bool bApproxStats;
+        double dfMin;
+        double dfMax;
+        double dfMean;
+        double dfStdDev;
+        GUInt64 nValidCount;
+    };
+    std::map<CPLString, Statistics> oMapMDArrayStatistics{};
 };
+//! @endcond
 
 /* ******************************************************************** */
 /*                           GDALPamDataset                             */
 /* ******************************************************************** */
 
+/** PAM dataset */
 class CPL_DLL GDALPamDataset : public GDALDataset
 {
     friend class GDALPamRasterBand;
 
-  protected:
-                GDALPamDataset(void);
+  private:
+    int IsPamFilenameAPotentialSiblingFile();
 
-    int         nPamFlags;
-    GDALDatasetPamInfo *psPam;
+  protected:
+
+                GDALPamDataset(void);
+//! @cond Doxygen_Suppress
+    int         nPamFlags = 0;
+    GDALDatasetPamInfo *psPam = nullptr;
+
+    virtual const char *_GetProjectionRef() override;
+    virtual const char *_GetGCPProjection() override;
+    virtual CPLErr _SetProjection( const char * pszProjection ) override;
+    virtual CPLErr _SetGCPs( int nGCPCount, const GDAL_GCP *pasGCPList,
+                    const char *pszGCPProjection ) override;
 
     virtual CPLXMLNode *SerializeToXML( const char *);
     virtual CPLErr      XMLInit( CPLXMLNode *, const char * );
-    
-    virtual CPLErr TryLoadXML();
+
+    virtual CPLErr TryLoadXML(char **papszSiblingFiles = nullptr);
     virtual CPLErr TrySaveXML();
 
-    CPLErr  TryLoadAux();
+    CPLErr  TryLoadAux(char **papszSiblingFiles = nullptr);
     CPLErr  TrySaveAux();
 
+    void SerializeMDArrayStatistics(CPLXMLNode* psDSTree);
+
     virtual const char *BuildPamFilename();
 
     void   PamInitialize();
@@ -132,55 +163,76 @@ class CPL_DLL GDALPamDataset : public GDALDataset
     const char *GetPhysicalFilename();
     void   SetSubdatasetName( const char *);
     const char *GetSubdatasetName();
+//! @endcond
 
   public:
-    virtual     ~GDALPamDataset();
+    ~GDALPamDataset() override;
 
-    virtual void FlushCache(void);
+    void FlushCache(void) override;
 
-    virtual const char *GetProjectionRef(void);
-    virtual CPLErr SetProjection( const char * );
+    const OGRSpatialReference* GetSpatialRef() const override;
+    CPLErr SetSpatialRef(const OGRSpatialReference* poSRS) override;
 
-    virtual CPLErr GetGeoTransform( double * );
-    virtual CPLErr SetGeoTransform( double * );
+    CPLErr GetGeoTransform( double * ) override;
+    CPLErr SetGeoTransform( double * ) override;
 
-    virtual int    GetGCPCount();
-    virtual const char *GetGCPProjection();
-    virtual const GDAL_GCP *GetGCPs();
-    virtual CPLErr SetGCPs( int nGCPCount, const GDAL_GCP *pasGCPList,
-                            const char *pszGCPProjection );
+    int GetGCPCount() override;
+    const OGRSpatialReference* GetGCPSpatialRef() const override;
+    const GDAL_GCP *GetGCPs() override;
+    using GDALDataset::SetGCPs;
+    CPLErr SetGCPs( int nGCPCount, const GDAL_GCP *pasGCPList,
+                    const OGRSpatialReference* poSRS ) override;
 
-    virtual CPLErr      SetMetadata( char ** papszMetadata,
-                                     const char * pszDomain = "" );
-    virtual CPLErr      SetMetadataItem( const char * pszName,
-                                         const char * pszValue,
-                                         const char * pszDomain = "" );
-    virtual char      **GetMetadata( const char * pszDomain = "" );
-    virtual const char *GetMetadataItem( const char * pszName,
-                                         const char * pszDomain = "" );
+    CPLErr SetMetadata( char ** papszMetadata,
+                        const char * pszDomain = "" ) override;
+    CPLErr SetMetadataItem( const char * pszName,
+                            const char * pszValue,
+                            const char * pszDomain = "" ) override;
+    char **GetMetadata( const char * pszDomain = "" ) override;
+    const char *GetMetadataItem( const char * pszName,
+                                 const char * pszDomain = "" ) override;
 
-    virtual char      **GetFileList(void);
+    char **GetFileList(void) override;
 
+    void ClearStatistics() override;
+
+//! @cond Doxygen_Suppress
     virtual CPLErr CloneInfo( GDALDataset *poSrcDS, int nCloneInfoFlags );
 
-    virtual CPLErr IBuildOverviews( const char *pszResampling, 
-                                    int nOverviews, int *panOverviewList, 
-                                    int nListBands, int *panBandList,
-                                    GDALProgressFunc pfnProgress, 
-                                    void * pProgressData );
+    CPLErr IBuildOverviews( const char *pszResampling,
+                            int nOverviews, int *panOverviewList,
+                            int nListBands, int *panBandList,
+                            GDALProgressFunc pfnProgress,
+                            void * pProgressData ) override;
 
+    bool GetMDArrayStatistics( const char* pszMDArrayId,
+                               bool *pbApprox,
+                               double *pdfMin, double *pdfMax,
+                               double *pdfMean, double *pdfStdDev,
+                               GUInt64 *pnValidCount );
+
+    void StoreMDArrayStatistics( const char* pszMDArrayId,
+                                 bool bApprox,
+                                 double dfMin, double dfMax,
+                                 double dfMean, double dfStdDev,
+                                 GUInt64 nValidCount );
 
     // "semi private" methods.
     void   MarkPamDirty() { nPamFlags |= GPF_DIRTY; }
     GDALDatasetPamInfo *GetPamInfo() { return psPam; }
     int    GetPamFlags() { return nPamFlags; }
     void   SetPamFlags(int nValue ) { nPamFlags = nValue; }
+//! @endcond
+
+  private:
+    CPL_DISALLOW_COPY_ASSIGN(GDALPamDataset)
 };
 
+//! @cond Doxygen_Suppress
 /* ==================================================================== */
 /*      GDALRasterBandPamInfo                                           */
 /*                                                                      */
-/*      We make these things a seperate structure of information        */
+/*      We make these things a separate structure of information        */
 /*      primarily so we can modify it without altering the size of      */
 /*      the GDALPamDataset.  It is an effort to reduce ABI churn for    */
 /*      driver plugins.                                                 */
@@ -197,7 +249,7 @@ typedef struct {
 
     char           *pszUnitType;
     char           **papszCategoryNames;
-    
+
     double         dfOffset;
     double         dfScale;
 
@@ -213,92 +265,106 @@ typedef struct {
 
     GDALRasterAttributeTable *poDefaultRAT;
 
+    bool           bOffsetSet;
+    bool           bScaleSet;
 } GDALRasterBandPamInfo;
-
+//! @endcond
 /* ******************************************************************** */
 /*                          GDALPamRasterBand                           */
 /* ******************************************************************** */
+
+/** PAM raster band */
 class CPL_DLL GDALPamRasterBand : public GDALRasterBand
 {
     friend class GDALPamDataset;
 
   protected:
-
+//! @cond Doxygen_Suppress
     virtual CPLXMLNode *SerializeToXML( const char *pszVRTPath );
     virtual CPLErr      XMLInit( CPLXMLNode *, const char * );
-    
+
     void   PamInitialize();
     void   PamClear();
 
-    GDALRasterBandPamInfo *psPam;
+    GDALRasterBandPamInfo *psPam = nullptr;
+//! @endcond
 
   public:
                 GDALPamRasterBand();
-    virtual     ~GDALPamRasterBand();
+//! @cond Doxygen_Suppress
+    explicit    GDALPamRasterBand(int bForceCachedIO);
+//! @endcond
+    ~GDALPamRasterBand() override;
 
-    virtual void        SetDescription( const char * );
+    void SetDescription( const char * ) override;
 
-    virtual CPLErr SetNoDataValue( double );
-    virtual double GetNoDataValue( int *pbSuccess = NULL );
+    CPLErr SetNoDataValue( double ) override;
+    double GetNoDataValue( int *pbSuccess = nullptr ) override;
+    CPLErr DeleteNoDataValue() override;
 
-    virtual CPLErr SetColorTable( GDALColorTable * ); 
-    virtual GDALColorTable *GetColorTable();
+    CPLErr SetColorTable( GDALColorTable * ) override;
+    GDALColorTable *GetColorTable() override;
 
-    virtual CPLErr SetColorInterpretation( GDALColorInterp );
-    virtual GDALColorInterp GetColorInterpretation();
+    CPLErr SetColorInterpretation( GDALColorInterp ) override;
+    GDALColorInterp GetColorInterpretation() override;
 
-    virtual const char *GetUnitType();
-    CPLErr SetUnitType( const char * ); 
+    const char *GetUnitType() override;
+    CPLErr SetUnitType( const char * ) override;
 
-    virtual char **GetCategoryNames();
-    virtual CPLErr SetCategoryNames( char ** );
+    char **GetCategoryNames() override;
+    CPLErr SetCategoryNames( char ** ) override;
 
-    virtual double GetOffset( int *pbSuccess = NULL );
-    CPLErr SetOffset( double );
-    virtual double GetScale( int *pbSuccess = NULL );
-    CPLErr SetScale( double );
+    double GetOffset( int *pbSuccess = nullptr ) override;
+    CPLErr SetOffset( double ) override;
+    double GetScale( int *pbSuccess = nullptr ) override;
+    CPLErr SetScale( double ) override;
 
-    virtual CPLErr  GetHistogram( double dfMin, double dfMax,
-                          int nBuckets, int * panHistogram,
-                          int bIncludeOutOfRange, int bApproxOK,
-                          GDALProgressFunc, void *pProgressData );
+    CPLErr GetHistogram( double dfMin, double dfMax,
+                         int nBuckets, GUIntBig * panHistogram,
+                         int bIncludeOutOfRange, int bApproxOK,
+                         GDALProgressFunc, void *pProgressData ) override;
 
-    virtual CPLErr GetDefaultHistogram( double *pdfMin, double *pdfMax,
-                                        int *pnBuckets, int ** ppanHistogram,
-                                        int bForce,
-                                        GDALProgressFunc, void *pProgressData);
+    CPLErr GetDefaultHistogram( double *pdfMin, double *pdfMax,
+                                int *pnBuckets, GUIntBig ** ppanHistogram,
+                                int bForce,
+                                GDALProgressFunc, void *pProgressData) override;
 
-    virtual CPLErr SetDefaultHistogram( double dfMin, double dfMax,
-                                        int nBuckets, int *panHistogram );
+    CPLErr SetDefaultHistogram( double dfMin, double dfMax,
+                                int nBuckets, GUIntBig *panHistogram ) override;
 
-    virtual CPLErr      SetMetadata( char ** papszMetadata,
-                                     const char * pszDomain = "" );
-    virtual CPLErr      SetMetadataItem( const char * pszName,
-                                         const char * pszValue,
-                                         const char * pszDomain = "" );
+    CPLErr SetMetadata( char ** papszMetadata,
+                        const char * pszDomain = "" ) override;
+    CPLErr SetMetadataItem( const char * pszName,
+                            const char * pszValue,
+                            const char * pszDomain = "" ) override;
 
-    virtual const GDALRasterAttributeTable *GetDefaultRAT();
-    virtual CPLErr SetDefaultRAT( const GDALRasterAttributeTable * );
+    GDALRasterAttributeTable *GetDefaultRAT() override;
+    CPLErr SetDefaultRAT( const GDALRasterAttributeTable * ) override;
 
-    // new in GDALPamRasterBand. 
+//! @cond Doxygen_Suppress
+    // new in GDALPamRasterBand.
     virtual CPLErr CloneInfo( GDALRasterBand *poSrcBand, int nCloneInfoFlags );
 
     // "semi private" methods.
     GDALRasterBandPamInfo *GetPamInfo() { return psPam; }
+//! @endcond
+  private:
+    CPL_DISALLOW_COPY_ASSIGN(GDALPamRasterBand)
 };
 
+//! @cond Doxygen_Suppress
 // These are mainly helper functions for internal use.
-int CPL_DLL PamParseHistogram( CPLXMLNode *psHistItem, 
-                               double *pdfMin, double *pdfMax, 
-                               int *pnBuckets, int **ppanHistogram, 
+int CPL_DLL PamParseHistogram( CPLXMLNode *psHistItem,
+                               double *pdfMin, double *pdfMax,
+                               int *pnBuckets, GUIntBig **ppanHistogram,
                                int *pbIncludeOutOfRange, int *pbApproxOK );
 CPLXMLNode CPL_DLL *
 PamFindMatchingHistogram( CPLXMLNode *psSavedHistograms,
-                          double dfMin, double dfMax, int nBuckets, 
+                          double dfMin, double dfMax, int nBuckets,
                           int bIncludeOutOfRange, int bApproxOK );
 CPLXMLNode CPL_DLL *
 PamHistogramToXMLTree( double dfMin, double dfMax,
-                       int nBuckets, int * panHistogram,
+                       int nBuckets, GUIntBig * panHistogram,
                        int bIncludeOutOfRange, int bApprox );
 
 // For managing the proxy file database.
@@ -307,4 +373,6 @@ const char CPL_DLL * PamAllocateProxy( const char * );
 const char CPL_DLL * PamDeallocateProxy( const char * );
 void CPL_DLL PamCleanProxyDB( void );
 
+//! @endcond
+
 #endif /* ndef GDAL_PAM_H_INCLUDED */
