@@ -1,5 +1,5 @@
 /******************************************************************************
- * $Id: ogr_api.h 7c7402f4226a5d821bfd6bf38af0e682c8720ba7 2020-04-27 13:06:52 +0200 Even Rouault $
+ * $Id: ogr_api.h 3798cbe48457b7127606931896549f26507469db 2021-04-09 15:04:16 +0200 Even Rouault $
  *
  * Project:  OpenGIS Simple Features Reference Implementation
  * Purpose:  C API for OGR Geometry, Feature, Layers, DataSource and drivers.
@@ -44,6 +44,9 @@
 #include "cpl_minixml.h"
 #include "ogr_core.h"
 
+#include <stdbool.h>
+#include <stddef.h>
+
 CPL_C_START
 
 /* -------------------------------------------------------------------- */
@@ -86,6 +89,8 @@ struct _CPLXMLNode;
 
 OGRErr CPL_DLL OGR_G_CreateFromWkb( const void*, OGRSpatialReferenceH,
                                     OGRGeometryH *, int );
+OGRErr CPL_DLL OGR_G_CreateFromWkbEx( const void*, OGRSpatialReferenceH,
+                                      OGRGeometryH *, size_t );
 OGRErr CPL_DLL OGR_G_CreateFromWkt( char **, OGRSpatialReferenceH,
                                     OGRGeometryH * );
 OGRErr CPL_DLL OGR_G_CreateFromFgf( const void*, OGRSpatialReferenceH,
@@ -124,6 +129,7 @@ OGRErr CPL_DLL OGR_G_ImportFromWkb( OGRGeometryH, const void*, int );
 OGRErr CPL_DLL OGR_G_ExportToWkb( OGRGeometryH, OGRwkbByteOrder, unsigned char*);
 OGRErr CPL_DLL OGR_G_ExportToIsoWkb( OGRGeometryH, OGRwkbByteOrder, unsigned char*);
 int    CPL_DLL OGR_G_WkbSize( OGRGeometryH hGeom );
+size_t CPL_DLL OGR_G_WkbSizeEx( OGRGeometryH hGeom );
 OGRErr CPL_DLL OGR_G_ImportFromWkt( OGRGeometryH, char ** );
 OGRErr CPL_DLL OGR_G_ExportToWkt( OGRGeometryH, char ** );
 OGRErr CPL_DLL OGR_G_ExportToIsoWkt( OGRGeometryH, char ** );
@@ -203,6 +209,7 @@ int    CPL_DLL OGR_G_IsEmpty( OGRGeometryH );
 int    CPL_DLL OGR_G_IsValid( OGRGeometryH );
 /*char    CPL_DLL *OGR_G_IsValidReason( OGRGeometryH );*/
 OGRGeometryH CPL_DLL OGR_G_MakeValid( OGRGeometryH ) CPL_WARN_UNUSED_RESULT;
+OGRGeometryH CPL_DLL OGR_G_Normalize( OGRGeometryH ) CPL_WARN_UNUSED_RESULT;
 int    CPL_DLL OGR_G_IsSimple( OGRGeometryH );
 int    CPL_DLL OGR_G_IsRing( OGRGeometryH );
 
@@ -292,6 +299,17 @@ int CPL_DLL OGRGetGenerate_DB2_V72_BYTE_ORDER(void);
 void CPL_DLL OGRSetNonLinearGeometriesEnabledFlag(int bFlag);
 int CPL_DLL OGRGetNonLinearGeometriesEnabledFlag(void);
 
+/** Opaque type for a prepared geometry */
+typedef struct _OGRPreparedGeometry * OGRPreparedGeometryH;
+
+int CPL_DLL OGRHasPreparedGeometrySupport(void);
+OGRPreparedGeometryH CPL_DLL OGRCreatePreparedGeometry( OGRGeometryH hGeom );
+void CPL_DLL OGRDestroyPreparedGeometry( OGRPreparedGeometryH hPreparedGeom );
+int CPL_DLL OGRPreparedGeometryIntersects( OGRPreparedGeometryH hPreparedGeom,
+                                           OGRGeometryH hOtherGeom );
+int CPL_DLL OGRPreparedGeometryContains( OGRPreparedGeometryH hPreparedGeom,
+                                         OGRGeometryH hOtherGeom );
+
 /* -------------------------------------------------------------------- */
 /*      Feature related (ogr_feature.h)                                 */
 /* -------------------------------------------------------------------- */
@@ -317,6 +335,9 @@ typedef void *OGRStyleTableH;
 #endif
 /** Opaque type for a geometry field definition (OGRGeomFieldDefn) */
 typedef struct OGRGeomFieldDefnHS *OGRGeomFieldDefnH;
+
+/** Opaque type for a field domain definition (OGRFieldDomain) */
+typedef struct OGRFieldDomainHS *OGRFieldDomainH;
 #endif /* DEFINE_OGRFeatureH */
 
 /* OGRFieldDefn */
@@ -326,6 +347,8 @@ void   CPL_DLL OGR_Fld_Destroy( OGRFieldDefnH );
 
 void   CPL_DLL OGR_Fld_SetName( OGRFieldDefnH, const char * );
 const char CPL_DLL *OGR_Fld_GetNameRef( OGRFieldDefnH );
+void   CPL_DLL OGR_Fld_SetAlternativeName( OGRFieldDefnH, const char * );
+const char CPL_DLL *OGR_Fld_GetAlternativeNameRef( OGRFieldDefnH );
 OGRFieldType CPL_DLL OGR_Fld_GetType( OGRFieldDefnH );
 void   CPL_DLL OGR_Fld_SetType( OGRFieldDefnH, OGRFieldType );
 OGRFieldSubType CPL_DLL OGR_Fld_GetSubType( OGRFieldDefnH );
@@ -342,9 +365,13 @@ int    CPL_DLL OGR_Fld_IsIgnored( OGRFieldDefnH hDefn );
 void   CPL_DLL OGR_Fld_SetIgnored( OGRFieldDefnH hDefn, int );
 int    CPL_DLL OGR_Fld_IsNullable( OGRFieldDefnH hDefn );
 void   CPL_DLL OGR_Fld_SetNullable( OGRFieldDefnH hDefn, int );
+int    CPL_DLL OGR_Fld_IsUnique( OGRFieldDefnH hDefn );
+void   CPL_DLL OGR_Fld_SetUnique( OGRFieldDefnH hDefn, int );
 const char CPL_DLL *OGR_Fld_GetDefault( OGRFieldDefnH hDefn );
 void   CPL_DLL OGR_Fld_SetDefault( OGRFieldDefnH hDefn, const char* );
 int    CPL_DLL OGR_Fld_IsDefaultDriverSpecific( OGRFieldDefnH hDefn );
+const char CPL_DLL* OGR_Fld_GetDomainName( OGRFieldDefnH hDefn );
+void   CPL_DLL OGR_Fld_SetDomainName( OGRFieldDefnH hDefn, const char* );
 
 const char CPL_DLL *OGR_GetFieldTypeName( OGRFieldType );
 const char CPL_DLL *OGR_GetFieldSubTypeName( OGRFieldSubType );
@@ -508,6 +535,44 @@ void   CPL_DLL OGR_F_FillUnsetWithDefault( OGRFeatureH hFeat,
                                            char** papszOptions );
 int    CPL_DLL OGR_F_Validate( OGRFeatureH, int nValidateFlags, int bEmitError );
 
+/* OGRFieldDomain */
+
+void CPL_DLL OGR_FldDomain_Destroy(OGRFieldDomainH);
+const char CPL_DLL* OGR_FldDomain_GetName(OGRFieldDomainH);
+const char CPL_DLL* OGR_FldDomain_GetDescription(OGRFieldDomainH);
+OGRFieldDomainType CPL_DLL OGR_FldDomain_GetDomainType(OGRFieldDomainH);
+OGRFieldType CPL_DLL OGR_FldDomain_GetFieldType(OGRFieldDomainH);
+OGRFieldSubType CPL_DLL OGR_FldDomain_GetFieldSubType(OGRFieldDomainH);
+OGRFieldDomainSplitPolicy CPL_DLL OGR_FldDomain_GetSplitPolicy(OGRFieldDomainH);
+void CPL_DLL OGR_FldDomain_SetSplitPolicy(OGRFieldDomainH, OGRFieldDomainSplitPolicy);
+OGRFieldDomainMergePolicy CPL_DLL OGR_FldDomain_GetMergePolicy(OGRFieldDomainH);
+void CPL_DLL OGR_FldDomain_SetMergePolicy(OGRFieldDomainH, OGRFieldDomainMergePolicy);
+
+OGRFieldDomainH CPL_DLL OGR_CodedFldDomain_Create(const char* pszName,
+                                                  const char* pszDescription,
+                                                  OGRFieldType eFieldType,
+                                                  OGRFieldSubType eFieldSubType,
+                                                  const OGRCodedValue* enumeration);
+const OGRCodedValue CPL_DLL* OGR_CodedFldDomain_GetEnumeration(OGRFieldDomainH);
+
+OGRFieldDomainH CPL_DLL OGR_RangeFldDomain_Create(const char* pszName,
+                                                  const char* pszDescription,
+                                                  OGRFieldType eFieldType,
+                                                  OGRFieldSubType eFieldSubType,
+                                                  const OGRField* psMin,
+                                                  bool bMinIsInclusive,
+                                                  const OGRField* psMax,
+                                                  bool bMaxIsInclusive);
+const OGRField CPL_DLL *OGR_RangeFldDomain_GetMin(OGRFieldDomainH, bool* pbIsInclusiveOut);
+const OGRField CPL_DLL *OGR_RangeFldDomain_GetMax(OGRFieldDomainH, bool* pbIsInclusiveOut);
+
+OGRFieldDomainH CPL_DLL OGR_GlobFldDomain_Create(const char* pszName,
+                                                  const char* pszDescription,
+                                                  OGRFieldType eFieldType,
+                                                  OGRFieldSubType eFieldSubType,
+                                                  const char* pszGlob);
+const char CPL_DLL *OGR_GlobFldDomain_GetGlob(OGRFieldDomainH);
+
 /* -------------------------------------------------------------------- */
 /*      ogrsf_frmts.h                                                   */
 /* -------------------------------------------------------------------- */
