@@ -1,5 +1,5 @@
 /******************************************************************************
- * $Id: vrtdataset.h 977f2da59e46852f49534361860bb92f9c6ad1fd 2019-03-14 21:46:14 +0100 Even Rouault $
+ * $Id: vrtdataset.h 8c8864a6008b3f523a8e96017e87482904fa665b 2020-12-18 18:51:04 +0100 Even Rouault $
  *
  * Project:  Virtual GDAL Datasets
  * Purpose:  Declaration of virtual gdal dataset classes.
@@ -7,7 +7,7 @@
  *
  ******************************************************************************
  * Copyright (c) 2001, Frank Warmerdam <warmerdam@pobox.com>
- * Copyright (c) 2007-2013, Even Rouault <even dot rouault at mines-paris dot org>
+ * Copyright (c) 2007-2013, Even Rouault <even dot rouault at spatialys.com>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -34,6 +34,7 @@
 #ifndef DOXYGEN_SKIP
 
 #include "cpl_hash_set.h"
+#include "cpl_minixml.h"
 #include "gdal_pam.h"
 #include "gdal_priv.h"
 #include "gdal_rat.h"
@@ -47,7 +48,7 @@
 int VRTApplyMetadata( CPLXMLNode *, GDALMajorObject * );
 CPLXMLNode *VRTSerializeMetadata( GDALMajorObject * );
 CPLErr GDALRegisterDefaultPixelFunc();
-
+CPLString VRTSerializeNoData(double dfVal, GDALDataType eDataType, int nPrecision);
 #if 0
 int VRTWarpedOverviewTransform( void *pTransformArg, int bDstToSrc,
                                 int nPointCount,
@@ -80,12 +81,23 @@ public:
     }
 
     ~VRTOverviewInfo() {
+        CloseDataset();
+    }
+
+    bool CloseDataset()
+    {
         if( poBand == nullptr )
-            /* do nothing */;
-        else if( poBand->GetDataset()->GetShared() )
-            GDALClose( /* (GDALDatasetH) */ poBand->GetDataset() );
+            return false;
+
+        GDALDataset* poDS = poBand->GetDataset();
+        // Nullify now, to prevent recursion in some cases !
+        poBand = nullptr;
+        if( poDS->GetShared() )
+            GDALClose( /* (GDALDatasetH) */ poDS );
         else
-            poBand->GetDataset()->Dereference();
+            poDS->Dereference();
+
+        return true;
     }
 };
 
@@ -154,43 +166,57 @@ template<class T> struct VRTFlushCacheStruct
 
 class VRTWarpedDataset;
 class VRTPansharpenedDataset;
+class VRTGroup;
 
-class CPL_DLL VRTDataset : public GDALDataset
+class CPL_DLL VRTDataset CPL_NON_FINAL: public GDALDataset
 {
     friend class VRTRasterBand;
     friend struct VRTFlushCacheStruct<VRTDataset>;
     friend struct VRTFlushCacheStruct<VRTWarpedDataset>;
     friend struct VRTFlushCacheStruct<VRTPansharpenedDataset>;
     friend class VRTSourcedRasterBand;
+    friend VRTDatasetH CPL_STDCALL VRTCreate(int nXSize, int nYSize);
 
     OGRSpatialReference* m_poSRS = nullptr;
 
-    int            m_bGeoTransformSet;
+    int            m_bGeoTransformSet = false;
     double         m_adfGeoTransform[6];
 
-    int            m_nGCPCount;
-    GDAL_GCP      *m_pasGCPList;
+    int            m_nGCPCount = 0;
+    GDAL_GCP      *m_pasGCPList = nullptr;
     OGRSpatialReference *m_poGCP_SRS = nullptr;
 
-    int            m_bNeedsFlush;
-    int            m_bWritable;
+    bool           m_bNeedsFlush = false;
+    bool           m_bWritable = true;
+    bool           m_bCanTakeRef = true;
 
-    char          *m_pszVRTPath;
+    char          *m_pszVRTPath = nullptr;
 
-    VRTRasterBand *m_poMaskBand;
+    VRTRasterBand *m_poMaskBand = nullptr;
 
-    int            m_bCompatibleForDatasetIO;
+    int            m_bCompatibleForDatasetIO = -1;
     int            CheckCompatibleForDatasetIO();
     void           ExpandProxyBands();
 
-    std::vector<GDALDataset*> m_apoOverviews;
-    std::vector<GDALDataset*> m_apoOverviewsBak;
-    char         **m_papszXMLVRTMetadata;
+    // Virtual (ie not materialized) overviews, created either implicitly
+    // when it is cheap to do it, or explicitly.
+    std::vector<GDALDataset*> m_apoOverviews{};
+    std::vector<GDALDataset*> m_apoOverviewsBak{};
+    CPLString           m_osOverviewResampling{};
+    std::vector<int>    m_anOverviewFactors{};
 
-    std::map<CPLString, GDALDataset*> m_oMapSharedSources;
+    char         **m_papszXMLVRTMetadata = nullptr;
+
+    std::map<CPLString, GDALDataset*> m_oMapSharedSources{};
+    std::shared_ptr<VRTGroup> m_poRootGroup{};
+
+    int            m_nRecursionCounter = 0;
 
     VRTRasterBand*      InitBand(const char* pszSubclass, int nBand,
                                  bool bAllowPansharpened);
+    static GDALDataset *OpenVRTProtocol( const char* pszSpec );
+    bool                AddVirtualOverview(int nOvFactor,
+                                           const char* pszResampling);
 
     CPL_DISALLOW_COPY_ASSIGN(VRTDataset)
 
@@ -201,10 +227,10 @@ class CPL_DLL VRTDataset : public GDALDataset
                  VRTDataset(int nXSize, int nYSize);
     virtual ~VRTDataset();
 
-    void          SetNeedsFlush() { m_bNeedsFlush = TRUE; }
+    void          SetNeedsFlush() { m_bNeedsFlush = true; }
     virtual void  FlushCache() override;
 
-    void SetWritable(int bWritableIn) { m_bWritable = bWritableIn; }
+    void SetWritable(int bWritableIn) { m_bWritable = CPL_TO_BOOL(bWritableIn); }
 
     virtual CPLErr          CreateMaskBand( int nFlags ) override;
     void SetMaskBand(VRTRasterBand* poMaskBand);
@@ -255,6 +281,8 @@ class CPL_DLL VRTDataset : public GDALDataset
     virtual CPLErr IBuildOverviews( const char *, int, int *,
                                     int, int *, GDALProgressFunc, void * ) override;
 
+    std::shared_ptr<GDALGroup> GetRootGroup() const override;
+
     /* Used by PDF driver for example */
     GDALDataset*        GetSingleSimpleSource();
     void                BuildVirtualOverviews();
@@ -268,6 +296,9 @@ class CPL_DLL VRTDataset : public GDALDataset
     static GDALDataset *Create( const char * pszName,
                                 int nXSize, int nYSize, int nBands,
                                 GDALDataType eType, char ** papszOptions );
+    static GDALDataset *CreateMultiDimensional( const char * pszFilename,
+                                                CSLConstList papszRootGroupOptions,
+                                                CSLConstList papszOptions );
     static CPLErr       Delete( const char * pszFilename );
 };
 
@@ -278,7 +309,7 @@ class CPL_DLL VRTDataset : public GDALDataset
 class GDALWarpOperation;
 class VRTWarpedRasterBand;
 
-class CPL_DLL VRTWarpedDataset : public VRTDataset
+class CPL_DLL VRTWarpedDataset final: public VRTDataset
 {
     int               m_nBlockXSize;
     int               m_nBlockYSize;
@@ -292,13 +323,13 @@ class CPL_DLL VRTWarpedDataset : public VRTDataset
 
     struct VerticalShiftGrid
     {
-        CPLString osVGrids;
-        int       bInverse;
-        double    dfToMeterSrc;
-        double    dfToMeterDest;
-        CPLStringList aosOptions;
+        CPLString osVGrids{};
+        int       bInverse = false;
+        double    dfToMeterSrc = 0.0;
+        double    dfToMeterDest = 0.0;
+        CPLStringList aosOptions{};
     };
-    std::vector<VerticalShiftGrid> m_aoVerticalShiftGrids;
+    std::vector<VerticalShiftGrid> m_aoVerticalShiftGrids{};
 
     friend class VRTWarpedRasterBand;
 
@@ -354,7 +385,7 @@ typedef enum
     GTAdjust_NoneWithoutWarning
 } GTAdjustment;
 
-class VRTPansharpenedDataset : public VRTDataset
+class VRTPansharpenedDataset final: public VRTDataset
 {
     friend class      VRTPansharpenedRasterBand;
 
@@ -362,9 +393,9 @@ class VRTPansharpenedDataset : public VRTDataset
     int               m_nBlockYSize;
     GDALPansharpenOperation* m_poPansharpener;
     VRTPansharpenedDataset* m_poMainDataset;
-    std::vector<VRTPansharpenedDataset*> m_apoOverviewDatasets;
+    std::vector<VRTPansharpenedDataset*> m_apoOverviewDatasets{};
     // Map from absolute to relative.
-    std::map<CPLString,CPLString> m_oMapToRelativeFilenames;
+    std::map<CPLString,CPLString> m_oMapToRelativeFilenames{};
 
     int               m_bLoadingOtherBands;
 
@@ -378,7 +409,7 @@ class VRTPansharpenedDataset : public VRTDataset
     GTAdjustment      m_eGTAdjustment;
     int               m_bNoDataDisabled;
 
-    std::vector<GDALDataset*> m_apoDatasetsToClose;
+    std::vector<GDALDataset*> m_apoDatasetsToClose{};
 
     CPL_DISALLOW_COPY_ASSIGN(VRTPansharpenedDataset)
 
@@ -425,9 +456,11 @@ public:
 /*      no raster access.  That is handled by derived classes.          */
 /************************************************************************/
 
-class CPL_DLL VRTRasterBand : public GDALRasterBand
+class CPL_DLL VRTRasterBand CPL_NON_FINAL: public GDALRasterBand
 {
   protected:
+    friend class VRTDataset;
+
     int            m_bIsMaskBand;
 
     int            m_bNoDataValueSet;
@@ -435,7 +468,7 @@ class CPL_DLL VRTRasterBand : public GDALRasterBand
     int            m_bHideNoDataValue;
     double         m_dfNoDataValue;
 
-    std::unique_ptr<GDALColorTable> m_poColorTable;
+    std::unique_ptr<GDALColorTable> m_poColorTable{};
 
     GDALColorInterp m_eColorInterp;
 
@@ -449,11 +482,11 @@ class CPL_DLL VRTRasterBand : public GDALRasterBand
 
     void           Initialize( int nXSize, int nYSize );
 
-    std::vector<VRTOverviewInfo> m_apoOverviews;
+    std::vector<VRTOverviewInfo> m_aoOverviewInfos{};
 
     VRTRasterBand *m_poMaskBand;
 
-    std::unique_ptr<GDALRasterAttributeTable> m_poRAT;
+    std::unique_ptr<GDALRasterAttributeTable> m_poRAT{};
 
     CPL_DISALLOW_COPY_ASSIGN(VRTRasterBand)
 
@@ -540,11 +573,11 @@ class CPL_DLL VRTRasterBand : public GDALRasterBand
 
 class VRTSimpleSource;
 
-class CPL_DLL VRTSourcedRasterBand : public VRTRasterBand
+class CPL_DLL VRTSourcedRasterBand CPL_NON_FINAL: public VRTRasterBand
 {
   private:
     int            m_nRecursionCounter;
-    CPLString      m_osLastLocationInfo;
+    CPLString      m_osLastLocationInfo{};
     char         **m_papszSourceList;
 
     bool           CanUseSourcesMinMaxImplementations();
@@ -563,6 +596,10 @@ class CPL_DLL VRTSourcedRasterBand : public VRTRasterBand
                    VRTSourcedRasterBand( GDALDataset *poDS, int nBand,
                                          GDALDataType eType,
                                          int nXSize, int nYSize );
+                   VRTSourcedRasterBand( GDALDataset *poDS, int nBand,
+                                         GDALDataType eType,
+                                         int nXSize, int nYSize,
+                                         int nBlockXSizeIn, int nBlockYSizeIn );
     virtual        ~VRTSourcedRasterBand();
 
     virtual CPLErr IRasterIO( GDALRWFlag, int, int, int, int,
@@ -656,7 +693,7 @@ class CPL_DLL VRTSourcedRasterBand : public VRTRasterBand
 /*                         VRTWarpedRasterBand                          */
 /************************************************************************/
 
-class CPL_DLL VRTWarpedRasterBand : public VRTRasterBand
+class CPL_DLL VRTWarpedRasterBand final: public VRTRasterBand
 {
   public:
                    VRTWarpedRasterBand( GDALDataset *poDS, int nBand,
@@ -675,7 +712,7 @@ class CPL_DLL VRTWarpedRasterBand : public VRTRasterBand
 /*                        VRTPansharpenedRasterBand                     */
 /************************************************************************/
 
-class VRTPansharpenedRasterBand : public VRTRasterBand
+class VRTPansharpenedRasterBand final: public VRTRasterBand
 {
     int               m_nIndexAsPansharpenedBand;
 
@@ -713,7 +750,7 @@ class VRTPansharpenedRasterBand : public VRTRasterBand
 
 class VRTDerivedRasterBandPrivateData;
 
-class CPL_DLL VRTDerivedRasterBand : public VRTSourcedRasterBand
+class CPL_DLL VRTDerivedRasterBand CPL_NON_FINAL: public VRTSourcedRasterBand
 {
     VRTDerivedRasterBandPrivateData* m_poPrivate;
     bool InitializePython();
@@ -774,7 +811,7 @@ class CPL_DLL VRTDerivedRasterBand : public VRTSourcedRasterBand
 
 class RawRasterBand;
 
-class CPL_DLL VRTRawRasterBand : public VRTRasterBand
+class CPL_DLL VRTRawRasterBand CPL_NON_FINAL: public VRTRasterBand
 {
     RawRasterBand  *m_poRawRaster;
 
@@ -809,6 +846,11 @@ class CPL_DLL VRTRawRasterBand : public VRTRasterBand
 
     void           ClearRawLink();
 
+    CPLVirtualMem *GetVirtualMemAuto( GDALRWFlag eRWFlag,
+                                      int *pnPixelSpace,
+                                      GIntBig *pnLineSpace,
+                                      char **papszOptions ) override;
+
     virtual void   GetFileList( char*** ppapszFileList, int *pnSize,
                                 int *pnMaxSize, CPLHashSet* hSetFiles ) override;
 };
@@ -817,7 +859,7 @@ class CPL_DLL VRTRawRasterBand : public VRTRasterBand
 /*                              VRTDriver                               */
 /************************************************************************/
 
-class VRTDriver : public GDALDriver
+class VRTDriver final: public GDALDriver
 {
     CPL_DISALLOW_COPY_ASSIGN(VRTDriver)
 
@@ -843,12 +885,13 @@ class VRTDriver : public GDALDriver
 /*                           VRTSimpleSource                            */
 /************************************************************************/
 
-class CPL_DLL VRTSimpleSource : public VRTSource
+class CPL_DLL VRTSimpleSource CPL_NON_FINAL: public VRTSource
 {
     CPL_DISALLOW_COPY_ASSIGN(VRTSimpleSource)
 
 protected:
     friend class VRTSourcedRasterBand;
+    friend class VRTDataset;
 
     GDALRasterBand      *m_poRasterBand;
 
@@ -866,16 +909,18 @@ protected:
     double              m_dfDstXSize;
     double              m_dfDstYSize;
 
-    int                 m_bNoDataSet;
-    double              m_dfNoDataValue;
-    CPLString           m_osResampling;
+    int                 m_bNoDataSet;       // should really be a member of VRTComplexSource as only taken into account by it
+    double              m_dfNoDataValue;    // same as above
+    CPLString           m_osResampling{};
 
     int                 m_nMaxValue;
 
     int                 m_bRelativeToVRTOri;
-    CPLString           m_osSourceFileNameOri;
+    CPLString           m_osSourceFileNameOri{};
     int                 m_nExplicitSharedStatus; // -1 unknown, 0 = unshared, 1 = shared
 
+    bool                m_bDropRefOnSrcBand;
+
     int                 NeedMaxValAdjustment() const;
 
 public:
@@ -892,7 +937,7 @@ public:
     void           SetSrcMaskBand( GDALRasterBand * );
     void           SetSrcWindow( double, double, double, double );
     void           SetDstWindow( double, double, double, double );
-    void           SetNoDataValue( double dfNoDataValue );
+    void           SetNoDataValue( double dfNoDataValue );  // should really be a member of VRTComplexSource
     const CPLString& GetResampling() const { return m_osResampling; }
     void           SetResampling( const char* pszResampling );
 
@@ -939,6 +984,7 @@ public:
     virtual CPLErr FlushCache() override;
 
     GDALRasterBand* GetBand();
+    GDALRasterBand* GetMaskBandMainBand() { return m_poMaskBandMainBand; }
     int             IsSameExceptBandNumber( VRTSimpleSource* poOtherSource );
     CPLErr          DatasetRasterIO(
                                GDALDataType eBandDataType,
@@ -959,7 +1005,7 @@ public:
 /*                          VRTAveragedSource                           */
 /************************************************************************/
 
-class VRTAveragedSource : public VRTSimpleSource
+class VRTAveragedSource final: public VRTSimpleSource
 {
     CPL_DISALLOW_COPY_ASSIGN(VRTAveragedSource)
 
@@ -1004,7 +1050,7 @@ typedef enum
     VRT_SCALING_EXPONENTIAL,
 } VRTComplexSourceScaling;
 
-class CPL_DLL VRTComplexSource : public VRTSimpleSource
+class CPL_DLL VRTComplexSource CPL_NON_FINAL: public VRTSimpleSource
 {
     CPL_DISALLOW_COPY_ASSIGN(VRTComplexSource)
     bool           AreValuesUnchanged() const;
@@ -1024,6 +1070,8 @@ protected:
 
     int            m_nColorTableComponent;
 
+    bool           m_bUseMaskBand = false;
+
     template <class WorkingDT>
     CPLErr          RasterIOInternal( int nReqXOff, int nReqYOff,
                                       int nReqXSize, int nReqYSize,
@@ -1070,6 +1118,8 @@ public:
 
     double  LookupValue( double dfInput );
 
+    void    SetUseMaskBand(bool bUseMaskBand) { m_bUseMaskBand = bUseMaskBand; }
+
     void    SetLinearScaling( double dfOffset, double dfScale );
     void    SetPowerScaling( double dfExponent,
                              double dfSrcMin,
@@ -1087,7 +1137,7 @@ public:
 /*                           VRTFilteredSource                          */
 /************************************************************************/
 
-class VRTFilteredSource : public VRTComplexSource
+class VRTFilteredSource CPL_NON_FINAL: public VRTComplexSource
 {
 private:
     int          IsTypeSupported( GDALDataType eTestType ) const;
@@ -1122,7 +1172,7 @@ public:
 /*                       VRTKernelFilteredSource                        */
 /************************************************************************/
 
-class VRTKernelFilteredSource : public VRTFilteredSource
+class VRTKernelFilteredSource CPL_NON_FINAL: public VRTFilteredSource
 {
     CPL_DISALLOW_COPY_ASSIGN(VRTKernelFilteredSource)
 
@@ -1154,7 +1204,7 @@ public:
 /*                       VRTAverageFilteredSource                       */
 /************************************************************************/
 
-class VRTAverageFilteredSource : public VRTKernelFilteredSource
+class VRTAverageFilteredSource final: public VRTKernelFilteredSource
 {
     CPL_DISALLOW_COPY_ASSIGN(VRTAverageFilteredSource)
 
@@ -1170,7 +1220,7 @@ public:
 /************************************************************************/
 /*                            VRTFuncSource                             */
 /************************************************************************/
-class VRTFuncSource : public VRTSource
+class VRTFuncSource final: public VRTSource
 {
     CPL_DISALLOW_COPY_ASSIGN(VRTFuncSource)
 
@@ -1213,6 +1263,527 @@ public:
     float               fNoDataValue;
 };
 
+/************************************************************************/
+/*                              VRTGroup                                */
+/************************************************************************/
+
+#ifdef TMPEXPORT
+#define TMP_CPL_DLL CPL_DLL
+#else
+#define TMP_CPL_DLL
+#endif
+
+class VRTMDArray;
+class VRTAttribute;
+class VRTDimension;
+
+class VRTGroup final: public GDALGroup
+{
+public:
+    struct Ref
+    {
+        VRTGroup* m_ptr;
+        explicit Ref(VRTGroup* ptr): m_ptr(ptr) {}
+        Ref(const Ref&) = delete;
+        Ref& operator=(const Ref&) = delete;
+    };
+
+private:
+    std::shared_ptr<Ref> m_poSharedRefRootGroup{};
+    std::weak_ptr<Ref> m_poWeakRefRootGroup{};
+    std::shared_ptr<Ref> m_poRefSelf{};
+
+    std::string m_osFilename{};
+    mutable bool m_bDirty = false;
+    std::string m_osVRTPath{};
+    std::map<std::string, std::shared_ptr<VRTGroup>> m_oMapGroups{};
+    std::map<std::string, std::shared_ptr<VRTMDArray>> m_oMapMDArrays{};
+    std::map<std::string, std::shared_ptr<VRTAttribute>> m_oMapAttributes{};
+    std::map<std::string, std::shared_ptr<VRTDimension>> m_oMapDimensions{};
+
+    std::shared_ptr<VRTGroup> OpenGroupInternal(const std::string& osName) const;
+    void SetRootGroupRef(const std::weak_ptr<Ref>& rgRef);
+    std::weak_ptr<Ref> GetRootGroupRef() const;
+
+public:
+
+    VRTGroup(const std::string& osParentName, const std::string& osName);
+    ~VRTGroup();
+
+    bool XMLInit(const std::shared_ptr<VRTGroup>& poRoot,
+                 const std::shared_ptr<VRTGroup>& poThisGroup,
+                 const CPLXMLNode* psNode,
+                 const char* pszVRTPath);
+
+    std::vector<std::string> GetMDArrayNames(CSLConstList papszOptions) const override;
+    std::shared_ptr<GDALMDArray> OpenMDArray(const std::string& osName,
+                                             CSLConstList papszOptions = nullptr) const override;
+
+    std::vector<std::string> GetGroupNames(CSLConstList papszOptions) const override;
+    std::shared_ptr<GDALGroup> OpenGroup(const std::string& osName,
+                                         CSLConstList) const override
+    {
+        return OpenGroupInternal(osName);
+    }
+
+    std::vector<std::shared_ptr<GDALDimension>> GetDimensions(CSLConstList) const override;
+
+    std::vector<std::shared_ptr<GDALAttribute>> GetAttributes(CSLConstList) const override;
+
+    std::shared_ptr<VRTDimension> GetDimension(const std::string& name) const {
+        auto oIter = m_oMapDimensions.find(name);
+        return oIter == m_oMapDimensions.end() ? nullptr : oIter->second;
+    }
+    std::shared_ptr<VRTDimension> GetDimensionFromFullName(const std::string& name,
+                                                           bool bEmitError) const;
+
+    std::shared_ptr<GDALGroup> CreateGroup(const std::string& osName,
+                                           CSLConstList papszOptions = nullptr) override;
+
+    std::shared_ptr<GDALDimension> CreateDimension(const std::string& osName,
+                                                           const std::string& osType,
+                                                           const std::string& osDirection,
+                                                           GUInt64 nSize,
+                                                           CSLConstList papszOptions = nullptr) override;
+
+    std::shared_ptr<GDALAttribute> CreateAttribute(
+        const std::string& osName,
+        const std::vector<GUInt64>& anDimensions,
+        const GDALExtendedDataType& oDataType,
+        CSLConstList papszOptions = nullptr) override;
+
+    std::shared_ptr<GDALMDArray> CreateMDArray(const std::string& osName,
+                                                       const std::vector<std::shared_ptr<GDALDimension>>& aoDimensions,
+                                                       const GDALExtendedDataType& oDataType,
+                                                       CSLConstList papszOptions) override;
+
+    void SetIsRootGroup();
+
+    const std::shared_ptr<Ref>& GetRef() const { return m_poRefSelf; }
+    VRTGroup* GetRootGroup() const;
+
+    const std::string& GetVRTPath() const { return m_osVRTPath; }
+    void SetDirty();
+    void SetFilename(const std::string& osFilename) { m_osFilename = osFilename; }
+    void Serialize() const;
+    CPLXMLNode* SerializeToXML( const char *pszVRTPathIn ) const;
+    void Serialize(CPLXMLNode* psParent, const char *pszVRTPathIn) const;
+};
+
+/************************************************************************/
+/*                            VRTDimension                              */
+/************************************************************************/
+
+class VRTDimension final: public GDALDimension
+{
+    std::weak_ptr<VRTGroup::Ref> m_poGroupRef;
+    std::string m_osIndexingVariableName;
+
+public:
+    VRTDimension(const std::shared_ptr<VRTGroup::Ref>& poGroupRef,
+                  const std::string& osParentName,
+                  const std::string& osName,
+                  const std::string& osType,
+                  const std::string& osDirection,
+                  GUInt64 nSize,
+                  const std::string& osIndexingVariableName):
+        GDALDimension(osParentName, osName, osType, osDirection, nSize),
+        m_poGroupRef(poGroupRef),
+        m_osIndexingVariableName(osIndexingVariableName)
+    {}
+
+    VRTGroup* GetGroup() const;
+
+    static std::shared_ptr<VRTDimension> Create(const std::shared_ptr<VRTGroup>& poThisGroup,
+                                                const std::string& osParentName,
+                                                const CPLXMLNode* psNode);
+
+    std::shared_ptr<GDALMDArray> GetIndexingVariable() const override;
+
+    bool SetIndexingVariable(std::shared_ptr<GDALMDArray> poIndexingVariable) override;
+
+    void Serialize(CPLXMLNode* psParent) const;
+};
+
+/************************************************************************/
+/*                            VRTAttribute                              */
+/************************************************************************/
+
+class VRTAttribute final: public GDALAttribute
+{
+    GDALExtendedDataType m_dt;
+    std::vector<std::string> m_aosList{};
+    std::vector<std::shared_ptr<GDALDimension>> m_dims{};
+
+protected:
+
+    bool IRead(const GUInt64* arrayStartIdx,
+                      const size_t* count,
+                      const GInt64* arrayStep,
+                      const GPtrDiff_t* bufferStride,
+                      const GDALExtendedDataType& bufferDataType,
+                      void* pDstBuffer) const override;
+
+    bool IWrite(const GUInt64* arrayStartIdx,
+                      const size_t* count,
+                      const GInt64* arrayStep,
+                      const GPtrDiff_t* bufferStride,
+                      const GDALExtendedDataType& bufferDataType,
+                      const void* pSrcBuffer) override;
+
+
+public:
+    VRTAttribute(const std::string& osParentName,
+                 const std::string& osName,
+                 const GDALExtendedDataType& dt,
+                 std::vector<std::string>&& aosList):
+        GDALAbstractMDArray(osParentName, osName),
+        GDALAttribute(osParentName, osName),
+        m_dt(dt),
+        m_aosList(std::move(aosList))
+    {
+        if( m_aosList.size() > 1 )
+        {
+            m_dims.emplace_back(std::make_shared<GDALDimension>(
+                std::string(), "dim",
+                std::string(), std::string(), m_aosList.size()));
+        }
+    }
+
+    VRTAttribute(const std::string& osParentName,
+                 const std::string& osName,
+                 GUInt64 nDim,
+                 const GDALExtendedDataType& dt):
+        GDALAbstractMDArray(osParentName, osName),
+        GDALAttribute(osParentName, osName),
+        m_dt(dt)
+    {
+        if( nDim != 0 )
+        {
+            m_dims.emplace_back(std::make_shared<GDALDimension>(
+                std::string(), "dim",
+                std::string(), std::string(), nDim));
+        }
+    }
+
+    static bool CreationCommonChecks(const std::string& osName,
+                                     const std::vector<GUInt64>& anDimensions,
+                                     const std::map<std::string, std::shared_ptr<VRTAttribute>>& oMapAttributes);
+
+    static std::shared_ptr<VRTAttribute> Create(const std::string& osParentName,
+                                                const CPLXMLNode* psNode);
+
+    const std::vector<std::shared_ptr<GDALDimension>>& GetDimensions() const override { return m_dims; }
+
+    const GDALExtendedDataType &GetDataType() const override { return m_dt; }
+
+    void Serialize(CPLXMLNode* psParent) const;
+};
+
+/************************************************************************/
+/*                          VRTMDArraySource                            */
+/************************************************************************/
+
+class VRTMDArraySource
+{
+public:
+    virtual ~VRTMDArraySource() = default;
+
+    virtual bool Read(const GUInt64* arrayStartIdx,
+                      const size_t* count,
+                      const GInt64* arrayStep,
+                      const GPtrDiff_t* bufferStride,
+                      const GDALExtendedDataType& bufferDataType,
+                      void* pDstBuffer) const = 0;
+
+    virtual void Serialize(CPLXMLNode* psParent, const char* pszVRTPath) const = 0;
+};
+
+/************************************************************************/
+/*                            VRTMDArray                                */
+/************************************************************************/
+
+class VRTMDArray final: public GDALMDArray
+{
+protected:
+    friend class VRTGroup; // for access to SetSelf()
+
+    std::weak_ptr<VRTGroup::Ref> m_poGroupRef;
+    std::string m_osVRTPath{};
+
+    GDALExtendedDataType m_dt;
+    std::vector<std::shared_ptr<GDALDimension>> m_dims;
+    std::map<std::string, std::shared_ptr<VRTAttribute>> m_oMapAttributes{};
+    std::vector<std::unique_ptr<VRTMDArraySource>> m_sources{};
+    std::shared_ptr<OGRSpatialReference> m_poSRS{};
+    std::vector<GByte> m_abyNoData{};
+    std::string m_osUnit{};
+    double m_dfScale = 1.0;
+    double m_dfOffset = 0.0;
+    bool m_bHasScale = false;
+    bool m_bHasOffset = false;
+
+    bool IRead(const GUInt64* arrayStartIdx,
+                      const size_t* count,
+                      const GInt64* arrayStep,
+                      const GPtrDiff_t* bufferStride,
+                      const GDALExtendedDataType& bufferDataType,
+                      void* pDstBuffer) const override;
+
+    void SetDirty();
+
+public:
+    VRTMDArray(const std::shared_ptr<VRTGroup::Ref>& poGroupRef,
+               const std::string& osParentName,
+               const std::string& osName,
+               const GDALExtendedDataType& dt,
+               std::vector<std::shared_ptr<GDALDimension>>&& dims,
+               std::map<std::string, std::shared_ptr<VRTAttribute>>&& oMapAttributes) :
+        GDALAbstractMDArray(osParentName, osName),
+        GDALMDArray(osParentName, osName),
+        m_poGroupRef(poGroupRef),
+        m_osVRTPath(poGroupRef->m_ptr->GetVRTPath()),
+        m_dt(dt),
+        m_dims(std::move(dims)),
+        m_oMapAttributes(std::move(oMapAttributes))
+    {
+    }
+
+    VRTMDArray(const std::shared_ptr<VRTGroup::Ref>& poGroupRef,
+               const std::string& osParentName,
+               const std::string& osName,
+               const std::vector<std::shared_ptr<GDALDimension>>& dims,
+               const GDALExtendedDataType& dt) :
+        GDALAbstractMDArray(osParentName, osName),
+        GDALMDArray(osParentName, osName),
+        m_poGroupRef(poGroupRef),
+        m_osVRTPath(poGroupRef->m_ptr->GetVRTPath()),
+        m_dt(dt),
+        m_dims(dims)
+    {
+    }
+
+    bool IsWritable() const override { return false; }
+
+    static std::shared_ptr<VRTMDArray> Create(const std::shared_ptr<VRTGroup>& poThisGroup,
+                                              const std::string& osParentName,
+                                              const CPLXMLNode* psNode);
+
+    const std::vector<std::shared_ptr<GDALDimension>>& GetDimensions() const override { return m_dims; }
+
+    std::vector<std::shared_ptr<GDALAttribute>> GetAttributes(CSLConstList) const override;
+
+    const GDALExtendedDataType &GetDataType() const override { return m_dt; }
+
+    bool SetSpatialRef(const OGRSpatialReference* poSRS) override;
+
+    std::shared_ptr<OGRSpatialReference> GetSpatialRef() const override { return m_poSRS; }
+
+    const void* GetRawNoDataValue() const override;
+
+    bool SetRawNoDataValue(const void* pRawNoData) override;
+
+    const std::string& GetUnit() const override { return m_osUnit; }
+
+    bool SetUnit(const std::string& osUnit) override {
+        m_osUnit = osUnit; return true; }
+
+    double GetOffset(bool* pbHasOffset, GDALDataType* peStorageType) const override
+    {
+        if( pbHasOffset) *pbHasOffset = m_bHasOffset;
+        if( peStorageType ) *peStorageType = GDT_Unknown;
+        return m_dfOffset;
+    }
+
+    double GetScale(bool* pbHasScale, GDALDataType* peStorageType) const override
+    {
+        if( pbHasScale) *pbHasScale = m_bHasScale;
+        if( peStorageType ) *peStorageType = GDT_Unknown;
+        return m_dfScale;
+    }
+
+    bool SetOffset(double dfOffset, GDALDataType /* eStorageType */ = GDT_Unknown) override
+    { SetDirty(); m_bHasOffset = true; m_dfOffset = dfOffset; return true; }
+
+    bool SetScale(double dfScale, GDALDataType /* eStorageType */ = GDT_Unknown) override
+    { SetDirty(); m_bHasScale = true; m_dfScale = dfScale; return true; }
+
+    void AddSource(std::unique_ptr<VRTMDArraySource>&& poSource);
+
+    std::shared_ptr<GDALAttribute> CreateAttribute(
+        const std::string& osName,
+        const std::vector<GUInt64>& anDimensions,
+        const GDALExtendedDataType& oDataType,
+        CSLConstList papszOptions = nullptr) override;
+
+    bool CopyFrom(GDALDataset* poSrcDS,
+                          const GDALMDArray* poSrcArray,
+                          bool bStrict,
+                          GUInt64& nCurCost,
+                          const GUInt64 nTotalCost,
+                          GDALProgressFunc pfnProgress,
+                          void * pProgressData) override;
+
+    void Serialize(CPLXMLNode* psParent, const char *pszVRTPathIn ) const;
+
+    VRTGroup* GetGroup() const;
+
+    const std::string& GetVRTPath() const { return m_osVRTPath; }
+};
+
+/************************************************************************/
+/*                       VRTMDArraySourceInlinedValues                  */
+/************************************************************************/
+
+class VRTMDArraySourceInlinedValues final: public VRTMDArraySource
+{
+    const VRTMDArray* m_poDstArray = nullptr;
+    bool m_bIsConstantValue;
+    std::vector<GUInt64> m_anOffset{};
+    std::vector<size_t> m_anCount{};
+    std::vector<GByte> m_abyValues{};
+    std::vector<size_t> m_anInlinedArrayStrideInBytes{};
+    GDALExtendedDataType m_dt;
+
+    VRTMDArraySourceInlinedValues(const VRTMDArraySourceInlinedValues&) = delete;
+    VRTMDArraySourceInlinedValues& operator=(const VRTMDArraySourceInlinedValues&) = delete;
+
+public:
+    VRTMDArraySourceInlinedValues(const VRTMDArray* poDstArray,
+                                  bool bIsConstantValue,
+                                  std::vector<GUInt64>&& anOffset,
+                                  std::vector<size_t>&& anCount,
+                                  std::vector<GByte>&& abyValues):
+        m_poDstArray(poDstArray),
+        m_bIsConstantValue(bIsConstantValue),
+        m_anOffset(std::move(anOffset)),
+        m_anCount(std::move(anCount)),
+        m_abyValues(std::move(abyValues)),
+        m_dt(poDstArray->GetDataType())
+    {
+        const auto nDims(poDstArray->GetDimensionCount());
+        m_anInlinedArrayStrideInBytes.resize(nDims);
+        if( !bIsConstantValue && nDims > 0 )
+        {
+            m_anInlinedArrayStrideInBytes.back() = poDstArray->GetDataType().GetSize();
+            for(size_t i = nDims - 1; i > 0; )
+            {
+                --i;
+                m_anInlinedArrayStrideInBytes[i] =
+                    m_anInlinedArrayStrideInBytes[i+1] * m_anCount[i+1];
+            }
+        }
+    }
+
+    ~VRTMDArraySourceInlinedValues();
+
+    static std::unique_ptr<VRTMDArraySourceInlinedValues> Create(
+                                                const VRTMDArray* poDstArray,
+                                                const CPLXMLNode* psNode);
+
+    bool Read(const GUInt64* arrayStartIdx,
+                      const size_t* count,
+                      const GInt64* arrayStep,
+                      const GPtrDiff_t* bufferStride,
+                      const GDALExtendedDataType& bufferDataType,
+                      void* pDstBuffer) const override;
+
+    void Serialize(CPLXMLNode* psParent, const char* pszVRTPath) const override;
+};
+
+/************************************************************************/
+/*                     VRTMDArraySourceRegularlySpaced                  */
+/************************************************************************/
+
+class VRTMDArraySourceRegularlySpaced final: public VRTMDArraySource
+{
+    double m_dfStart;
+    double m_dfIncrement;
+
+public:
+    VRTMDArraySourceRegularlySpaced(
+                 double dfStart, double dfIncrement):
+        m_dfStart(dfStart),
+        m_dfIncrement(dfIncrement)
+    {
+    }
+
+    bool Read(const GUInt64* arrayStartIdx,
+                      const size_t* count,
+                      const GInt64* arrayStep,
+                      const GPtrDiff_t* bufferStride,
+                      const GDALExtendedDataType& bufferDataType,
+                      void* pDstBuffer) const override;
+
+    void Serialize(CPLXMLNode* psParent, const char* pszVRTPath) const override;
+};
+
+/************************************************************************/
+/*                       VRTMDArraySourceFromArray                      */
+/************************************************************************/
+
+class VRTMDArraySourceFromArray final: public VRTMDArraySource
+{
+    const VRTMDArray* m_poDstArray = nullptr;
+    bool m_bRelativeToVRTSet = false;
+    bool m_bRelativeToVRT = false;
+    std::string m_osFilename{};
+    std::string m_osArray{};
+    std::string m_osBand{};
+    std::vector<int> m_anTransposedAxis{};
+    std::string m_osViewExpr{};
+    std::vector<GUInt64> m_anSrcOffset{};
+    mutable std::vector<GUInt64> m_anCount{};
+    std::vector<GUInt64> m_anStep{};
+    std::vector<GUInt64> m_anDstOffset{};
+
+    VRTMDArraySourceFromArray(const VRTMDArraySourceFromArray&) = delete;
+    VRTMDArraySourceFromArray& operator=(const VRTMDArraySourceFromArray&) = delete;
+
+public:
+    VRTMDArraySourceFromArray(const VRTMDArray* poDstArray,
+                              bool bRelativeToVRTSet,
+                              bool bRelativeToVRT,
+                              const std::string& osFilename,
+                              const std::string& osArray,
+                              const std::string& osBand,
+                              std::vector<int>&& anTransposedAxis,
+                              const std::string& osViewExpr,
+                              std::vector<GUInt64>&& anSrcOffset,
+                              std::vector<GUInt64>&& anCount,
+                              std::vector<GUInt64>&& anStep,
+                              std::vector<GUInt64>&& anDstOffset):
+        m_poDstArray(poDstArray),
+        m_bRelativeToVRTSet(bRelativeToVRTSet),
+        m_bRelativeToVRT(bRelativeToVRT),
+        m_osFilename(osFilename),
+        m_osArray(osArray),
+        m_osBand(osBand),
+        m_anTransposedAxis(std::move(anTransposedAxis)),
+        m_osViewExpr(osViewExpr),
+        m_anSrcOffset(std::move(anSrcOffset)),
+        m_anCount(std::move(anCount)),
+        m_anStep(std::move(anStep)),
+        m_anDstOffset(std::move(anDstOffset))
+    {
+    }
+
+    ~VRTMDArraySourceFromArray() override;
+
+    static std::unique_ptr<VRTMDArraySourceFromArray> Create(
+                                                const VRTMDArray* poDstArray,
+                                                const CPLXMLNode* psNode);
+
+    bool Read(const GUInt64* arrayStartIdx,
+                      const size_t* count,
+                      const GInt64* arrayStep,
+                      const GPtrDiff_t* bufferStride,
+                      const GDALExtendedDataType& bufferDataType,
+                      void* pDstBuffer) const override;
+
+    void Serialize(CPLXMLNode* psParent, const char* pszVRTPath) const override;
+};
+
 #endif /* #ifndef DOXYGEN_SKIP */
 
 #endif /* ndef VIRTUALDATASET_H_INCLUDED */
